//
//  CollapsibleView.swift
//  SplatoonWhatsPlaying
//

import Cocoa
import Quartz

class CollapsibleViewController: NSViewController {
    @IBOutlet weak var titleTextField:NSTextField?
    @IBOutlet weak var hideButton:NSButton?
    @IBOutlet weak var headerView:NSView?
    
    var isCollapsed = false
    var closingConstraint:NSLayoutConstraint?
    
    // Defining init() below means the superclass's initializers are not inherited
    override init?(nibName: String?, bundle: Bundle?) {
        super.init(nibName: nibName, bundle: bundle)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    convenience init() {
        self.init(nibName: "CollapsibleView", bundle: nil)!
        // Ensure view is loaded now, so that the outlets are set
        _ = view
    }
    
    override var title:String? {
        get {
            return titleTextField?.stringValue
        }
        set {
            titleTextField?.stringValue = newValue ?? ""
        }
    }

    var collapsibleView:NSView? {
        willSet {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
            if newValue == nil {
                collapsibleView?.removeFromSuperview()
            }
            else if collapsibleView != newValue {
                // headerView should be connected in Interface Builder and should never be nil
                assert(headerView != nil)
                
                // Unclear if changing the view works - do we need to get rid of the constraints we added previously?
                collapsibleView?.removeFromSuperview()
                newValue!.removeFromSuperview()
                view.addSubview(newValue!)
                
                // Supports Core Animation
                newValue!.wantsLayer = true
                
                view.addConstraints(NSLayoutConstraint.constraints(withVisualFormat: "H:|[collapsibleView]|", options: NSLayoutFormatOptions(rawValue: 0), metrics: nil, views: ["collapsibleView":newValue!]))
                view.addConstraints(NSLayoutConstraint.constraints(withVisualFormat: "V:[headerView][collapsibleView]", options: NSLayoutFormatOptions(rawValue: 0), metrics: nil, views: ["collapsibleView":newValue!, "headerView":headerView!]))
                view.addConstraints(NSLayoutConstraint.constraints(withVisualFormat: "V:[collapsibleView]-(0@600)-|", options: NSLayoutFormatOptions(rawValue: 0), metrics: nil, views: ["collapsibleView":newValue!]))
            }
        }
    }
    
    // The Hide button was clicked
    @IBAction func toggleHidden(_ sender: AnyObject) {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        // User requests un-hiding
        if isCollapsed {
            // Should be set the first time toggleHidden is called
            assert(closingConstraint != nil)
            
            NSAnimationContext.runAnimationGroup({ (context:NSAnimationContext) -> Void in context.timingFunction = CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut); self.closingConstraint?.animator().constant += (self.collapsibleView?.frame.size.height)!; self.hideButton?.title = "Hide" }, completionHandler: { () -> Void in self.view.removeConstraint(self.closingConstraint!); self.isCollapsed = false })
        }
        // User requests hiding
        else {
            // headerView should be connected in Interface Builder and should never be nil
            assert(headerView != nil)
            
            let distanceFromHeaderToBottom =  NSMinY(headerView!.frame) - NSMinY(view.bounds)
            
            // If the hide button is pressed repeatedly, we can re-use the constraint
            if closingConstraint == nil {
                closingConstraint = NSLayoutConstraint(item: self.view, attribute: .bottom, relatedBy: .equal, toItem: headerView!, attribute: .bottom, multiplier: 1, constant: distanceFromHeaderToBottom)
            }
            
            closingConstraint!.constant = distanceFromHeaderToBottom
            view.addConstraint(closingConstraint!)

            // Animate the closing constraint to 0, causing the bottom of the header to be flush with the bottom of the overall disclosure view.
            NSAnimationContext.runAnimationGroup({ (context:NSAnimationContext) -> Void in context.timingFunction = CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut); self.closingConstraint?.animator().constant = 0; self.hideButton?.title = "Show" }, completionHandler: { () -> Void in self.isCollapsed = true })
        }
        
    }
}
