//
//  TodayViewController.swift
//  Widget
//

import Cocoa
import NotificationCenter

class TodayViewController: NSViewController, NCWidgetProviding {
    var topStack: NSStackView!
    
    var times: [String]!
    var regular_maps: [[String]]!
    var ranked_maps: [[String]]!
    var ranked_rules: [String]!
    var is_splatfest: Bool!
    
    var collapsibleViewControllers: [CollapsibleViewController]!
    
    override func loadView() {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        view = NSView(frame: NSZeroRect)
        view.translatesAutoresizingMaskIntoConstraints = false
        
        topStack = NSStackView(frame: NSZeroRect)
        topStack.translatesAutoresizingMaskIntoConstraints = false
        topStack.wantsLayer = true
        topStack.orientation = .vertical;
        
        topStack.alignment = NSLayoutAttribute.left
        topStack.spacing = 0

        topStack.setHuggingPriority(NSLayoutPriorityDefaultHigh, for: .horizontal)
        topStack.setHuggingPriority(NSLayoutPriorityDefaultHigh, for: .vertical)
        
        view.addSubview(topStack)
        view.addConstraints(NSLayoutConstraint.constraints(withVisualFormat: "H:|[topStack]-|", options: NSLayoutFormatOptions(rawValue: 0), metrics: nil, views: ["topStack":topStack]))
        view.addConstraints(NSLayoutConstraint.constraints(withVisualFormat: "V:|[topStack]|", options: NSLayoutFormatOptions(rawValue: 0), metrics: nil, views: ["topStack":topStack]))
        //view.window?.visualizeConstraints(view.constraints)
    }
    
    func updateSchedule () -> Date? {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        var nextUpdate: Date?
        
        top: if let jsonData = downloadUrl("http://splatoon.ink/schedule.json") {
            times = []
            regular_maps = []
            ranked_maps = []
            ranked_rules = []
            is_splatfest = false

            var startTimes: [Date] = []
            var endTimes: [Date] = []
            
            let json = JSON(data:jsonData.data(using: String.Encoding.utf8)!)
            
            if let is_sf = json["splatfest"].number, let schedule = json["schedule"].array {
                is_splatfest = (is_sf == true)

                for i in 0..<(schedule.count) {
                    if is_splatfest! {
                        let regular_m = schedule[i]["regular"]["maps"].array
                        
                        let st = schedule[i]["startTime"].number
                        let et = schedule[i]["endTime"].number
                        
                        // During splatfest, currently supports only configuration of three regular maps
                        if regular_m == nil || regular_m!.count != 3 || st == nil || et == nil {
                            break top
                        }

                        regular_maps.insert([], at: i)
                        
                        for j in 0..<(regular_m!.count) {
                            let regular_name = regular_m![j]["name"]["en"].string
                            
                            if regular_name == nil {
                                break top
                            }
                            
                            regular_maps[i].insert(regular_name!, at:j)
                        }
                        
                        startTimes.insert(Date(timeIntervalSince1970: (st!.doubleValue / 1000.0)), at:i)
                        endTimes.insert(Date(timeIntervalSince1970: (et!.doubleValue / 1000.0)), at:i)
                        
                    } // is_splatfest == true
                    else {
                        let regular_m = schedule[i]["regular"]["maps"].array
                        let ranked_m = schedule[i]["ranked"]["maps"].array
                        let ranked_r = schedule[i]["ranked"]["rules"]["en"].string
                        
                        let st = schedule[i]["startTime"].number
                        let et = schedule[i]["endTime"].number
                        
                        // Outside splatfest, currently supports only configuration of two regular maps, two ranked maps
                        if regular_m == nil || ranked_m == nil || regular_m!.count != 2 || ranked_m!.count != 2 || ranked_r == nil || st == nil || et == nil {
                            break top
                        }
                        
                        regular_maps.insert([], at: i)
                        ranked_maps.insert([], at: i)
                        
                        for j in 0..<(regular_m!.count) {
                            let regular_name = regular_m![j]["name"]["en"].string
                            let ranked_name = ranked_m![j]["name"]["en"].string
                            
                            if regular_name == nil || ranked_name == nil {
                                break top
                            }
                            
                            regular_maps[i].insert(regular_name!, at:j)
                            ranked_maps[i].insert(ranked_name!, at:j)
                        }
                        
                        ranked_rules.insert(ranked_r!, at:i)
                        
                        startTimes.insert(Date(timeIntervalSince1970: (st!.doubleValue / 1000.0)), at:i)
                        endTimes.insert(Date(timeIntervalSince1970: (et!.doubleValue / 1000.0)), at:i)
                    } // is_splatfest == false
                }
                
                // The json was fully parsed. Under normal conditions there were three timeblocks in the schedule, though could be as few as one (during or immediately prior to splatfest)
                
                let timeFormat = DateFormatter()
                timeFormat.dateFormat = "HH:mm"
                
                for i in 0..<startTimes.count {
                    times.insert("\(timeFormat.string(from: startTimes[i])) to \(timeFormat.string(from: endTimes[i]))", at: i)
                }

                var boardChange: Date?
                var refetch: Date?
                
                // The stages change on-time. The new schedules are uploaded about three minutes later, because Nintendo needs to post the new schedules and then splatoon.ink needs to grab and parse them. We will fetch them five minutes after to be safe.
                refetch = endTimes[0].addingTimeInterval(5*60)
                
                // Assume the boards will change at XX:00. So from XX:00 to XX:05 there are (usually) TWO maps to show the user. So at XX:00 we update the display (boardChange is XX:00) and at XX:05 we fetch a new schedule.
                while isInThePast(endTimes[0]) {
                    // The first map has ended, so remove it

                    // There may be a glitch in splatoon.ink, so wait at least one more minute before trying again
                    let fiveMinutesFromMapChange = endTimes[0].addingTimeInterval(5*60)
                    let oneMinuteFromNow = Date().addingTimeInterval(1*60)
                    refetch = (oneMinuteFromNow as NSDate).laterDate(fiveMinutesFromMapChange)
                    
                    // Get rid of stale map - might be the only map in which case we don't display anything until we refetch
                    times.removeFirst()
                    regular_maps.removeFirst()
                    ranked_maps.removeFirst()
                    ranked_rules.removeFirst()
                    startTimes.removeFirst()
                    endTimes.removeFirst()
                    
                    // Unless we nuked the only map, we are in the first few minutes of the first (of usually two) remaining maps
                }
                
                assert(isInTheFuture(refetch!))

                if endTimes.count > 0 {
                    boardChange = endTimes[0]
                    assert(isInTheFuture(boardChange!))

                    nextUpdate = (boardChange! as NSDate).earlierDate(refetch!)                }
                else {
                    nextUpdate = refetch!
                }
                

            } // if let is_sf = ..., let schedule = ...
        } // if let jsonData = downloadUrl...
        
        // If the json is not parsed properly, try again one minute from now
        if nextUpdate == nil {
            nextUpdate = Date().addingTimeInterval(1*60)
        }
        
        return nextUpdate
    } // func updateSchedule
    
    func isInThePast (_ date: Date) -> Bool { return date.compare(Date()) == .orderedAscending }
    
    func isInTheFuture (_ date: Date) -> Bool { return Date().compare(date) == .orderedAscending }
    
    func widgetPerformUpdate(completionHandler: @escaping ((NCUpdateResult) -> Swift.Void)) {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        // We update data in the background, so we probably have new data
        completionHandler(.newData)
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") didn't crash in \(#function)")
    }

    func updateViews() {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        // If we replace all of our views with new views we get a jarring visual effect - try to avoid if possible
        if canRecycleViews() {
            recycleViews()
        }
        else {
            installNewViews ()
        }
    }
    
    func canRecycleViews() -> Bool {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        if is_splatfest! || collapsibleViewControllers == nil || times.count != collapsibleViewControllers.count {
            // We don't support recycling old time blocks into Splatfest time blocks
            // We also don't support recycling when we have too many or too few time blocks
            
            return false
        }
        
        for cv in collapsibleViewControllers {
            let v = cv.collapsibleView!
            if v is SplatfestThreeRegularView {
                // We don't support recycling Splatfest time blocks, since they only happen once a month

                return false
            }
        }
        
        return true
    }
    
    func recycleViews() {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        for i in 0..<times.count {
            let cv = collapsibleViewControllers[i]
            cv.title = times[i]

            let v = cv.collapsibleView! as! TwoRegularTwoRankedView
            if (regular_maps[i].count == 2 && ranked_maps[i].count == 2) {
                v.recycle(regular_maps[i], ranked_m: ranked_maps[i], ranked_r: ranked_rules[i])
            }

        }
        
    }

    func installNewViews() {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        collapsibleViewControllers = []
        var views:[NSView] = []
        
        //print("updating Views")
        for i in 0..<times.count {
            let cv = CollapsibleViewController()
            var v:NSView?
            
            if (is_splatfest! == false && regular_maps[i].count == 2 && ranked_maps[i].count == 2) {
                v = TwoRegularTwoRankedView(regular_m: regular_maps[i], ranked_m: ranked_maps[i], ranked_r: ranked_rules[i])
            }
            else if (is_splatfest! == true && regular_maps[i].count == 3) {
                v = SplatfestThreeRegularView(regular_m: regular_maps[i])
            }
            
            cv.title = times[i]
            cv.collapsibleView = v
            
            collapsibleViewControllers.insert(cv, at: i)
            views.insert(cv.view, at: i)
            //print("inserted new collapsible view with title \(times[i])")
        }
        
        //print("replacing views in topStack")
        topStack.setViews(views, in: .center)
    }
    
    func installBlankViews() {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        collapsibleViewControllers = []
        var views:[NSView] = []
        
        //print("updating Views")
        for i in 0...2 {
            let cv = CollapsibleViewController()
            
            cv.title = ""
            cv.collapsibleView = TwoRegularTwoRankedView()
            
            collapsibleViewControllers.insert(cv, at: i)
            views.insert(cv.view, at: i)
            //print("inserted new collapsible view with title \(times[i])")
        }
        
        //print("replacing views in topStack")
        topStack.setViews(views, in: .center)
    }

    
    override func viewWillAppear() {
        super.viewWillAppear()
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        installBlankViews()
        
        let oneTenthSecondFromNow = Date().addingTimeInterval(0.1)
        scheduleTimer(oneTenthSecondFromNow)
    }
    
    func scheduleTimer (_ fireTime: Date?) {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        var waitTime: TimeInterval
        
        //let format = NSDateFormatter()
        //format.dateFormat = "yyyy-MM-dd HH:mm:ss"
        //print("boardChange \(format.stringFromDate(boardChange!)) and refetch \(format.stringFromDate(refetch!))")
        if fireTime != nil && isInTheFuture(fireTime!) {
            waitTime = fireTime!.timeIntervalSinceNow
        }
        else {
            // Some kind of error occurred in updateSchedule, let's just try again in five minutes
            waitTime = 300.0
        }
        
        //print("Timer scheduled for \(waitTime) seconds from now")
        Timer.scheduledTimer(timeInterval: waitTime, target: self, selector: #selector(TodayViewController.timerCallback), userInfo: nil, repeats: false)
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") didn't crash in \(#function)")
    }
    
    func timerCallback () {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        //print("Timer called")
        let nextUpdate = updateSchedule()
        updateViews()
        scheduleTimer(nextUpdate)
    }
    
    
    func downloadUrl (_ url: String) -> String? {
NSLog("\(Thread.isMainThread ? "(Main thread)" : "(Not main thread)") \(#function)")
        let config = URLSessionConfiguration.default
        let session = URLSession(configuration: config, delegate: nil, delegateQueue: nil)
        var semaphore: DispatchSemaphore
        semaphore = DispatchSemaphore(value: 0)
        
        var html: String?
        session.dataTask(with: URL(string: url)!, completionHandler: { data, response, error in html = NSString(data: data!, encoding: String.Encoding.utf8.rawValue) as? String ; semaphore.signal() } ).resume()
        
        _ = semaphore.wait(timeout: DispatchTime.distantFuture)
        return html
    }
}
